> module Core (module Core, Lab(..), ppLab, UnOp(..), BinOp(..), Prim(..))
> where
> import Data.List (sort)
>
> import AbstractSyntax (Lab(..), ppLab, UnOp(..), BinOp(..), Prim(..))
> import qualified AbstractSyntax as AS
> import Message

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Expressions}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Var   =  String
> type Con   =  String

> data Expr  
>   =  Bool     Bool                            -- truth value
>   |  If       Expr Expr Expr                  -- alternative
>   |  UnOp     UnOp Expr                       -- unary operator (on the naturals)
>   |  BinOp    Expr BinOp Expr                 -- binary operator (on the naturals)
>   |  Nat      Integer                         -- numeral
>   |  String   String                          -- string literal
>   |  Type                                     -- type argument marker
>   |  Var      Var                             -- variable
>   |  Let      Pat  Expr Expr                  -- local declaration
>   |  Fun      Pat  Expr                       -- abstraction
>   |  App      Expr Expr                       -- application
>   |  Rec      Pat  Expr                       -- recursion
>   |  Record   [Row Expr]                      -- record construction
>   |  Sel      Expr Lab                        -- record selection
>   |  Con      Con Expr                        -- constructor application
>   |  Case     Expr [(Pat, Expr)]              -- case analysis

>--   |  TyApp    Expr Type                     -- type application
>   |  Seq      Expr Expr                       -- sequence: e1 ; e2 ~= let _ = e1 in e2
>   |  Assign   Expr Expr                       -- assignment
>   |  Ref      Expr                            -- allocation
>   |  Deref    Expr                            -- dereferencing
>   |  LetCC    Var  Expr                       -- let current continuation
>   |  Continue Expr Expr                       -- apply a continuation
>   |  Throw    Expr                            -- throw an exception
>   |  Try      Expr [(Pat, Expr)]              -- catch an exception
>   |  Call     Prim Expr
>--   |  Write    Expr                            -- output a string
>--   |  Show     Expr                            -- convert to a string
>   |  While    Expr Expr                       -- while-loop

First-class environments.

>   |  Push    Expr
>   |  Pop     Expr
>   |  LetOpen Expr Expr                        -- let open
>   |  Struct
>   |  Lookup   Expr Var

Objects and classes.

>  |  Class Expr
>  |  New Expr
>  |  Empty
>  |  Append Expr Expr
>  |  Method Var Expr
>  |  Inherit Expr
>  |  InheritIn Var Expr Expr

Arrays.

>   |  ArrFun Expr Var Expr
>   |  ArrList [Expr]
>   |  Subscript Expr Expr

Fr Evaluate, um Frames in Expr berfhren zu knnen und dann
konsistent ausgeben zu knnen:

> --  |  Hole
>      deriving (Show)

> unit       =  Record []

Records are sorted internally, but only \emph{after} evaluation
(so that the evaluation order is the textual order).

> append Empty e  =  e
> append e Empty  =  e
> append e1 e2    =  Append e1 e2

> instance Pretty Expr

> ppBool False  =  text "false"
> ppBool True   =  text "true"

> ppExpr (Bool b) 
>   =  ppBool b
> ppExpr (If e1 e2 e3)
>   =  group (text "if " <> 
>             block 3 (ppExpr e1 <+> block 5 (text "then " <> ppExpr e2) <+>
>                                    block 5 (text "else " <> ppExpr e3)))
> ppExpr (UnOp op e)
>   =  parens (group (pretty op <> text " " <> block 2 (ppExpr e)))
> ppExpr (BinOp e1 op e2)
>   =  parens (group (ppExpr e1 <> text " " <> block 2 (pretty op <+> ppExpr e2)))
> ppExpr (Nat n) 
>   =  pretty n
> ppExpr (String s) 
>   =  pretty s
> ppExpr Type
>   =  text "<>"
> ppExpr (Var x) 
>   =  text x
> ppExpr (Let p1 e1 e)
>   =  group (text "let " <> block 4 (ppPat p1 <+> text "=" <+> ppExpr e1) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e)))
> ppExpr (Fun p e)
>   =  parens (align "\\ " (ppPat p <> text " =>" <+> ppExpr e))
> ppExpr (App e1 e2)
>   =  parens (group (ppExpr e1 <> text " @" <+> ppExpr e2))
> ppExpr (Rec p e)
>   =  parens (align "rec " (ppPat p <> text " => " <+> ppExpr e))
> ppExpr (Sel e l)
>   =  ppExpr e <> text "." <> ppLab l
> ppExpr (Record rs)
>   =  parens (group (intersperse (char ',' <> sp) (map (ppRow ppExpr) rs)))
> ppExpr (Con c e) 
>   =  parens (text c <+> ppExpr e)
> ppExpr (Case e rs)
>   =  group (text "case " <> ppExpr e <+> text "of " <>
>        intersperse (sp <> text " | ") (map ppRule rs) <+> text "end")

> ppExpr (Seq e1 e2)
>   =  group (ppExpr e1 <> text ";" <+> ppExpr e2)
> ppExpr (Assign e1 e2)
>   =  group (ppExpr e1 <> text " :=" <+> ppExpr e2)
> ppExpr (Ref e)
>   =  group (text "ref" <+> ppExpr e)
> ppExpr (Deref e)
>   =  group (text "!" <> ppExpr e)
> ppExpr (LetCC x e)
>   =  group (text "letcc " <> block 6 (text x) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e)))
> ppExpr (Continue e1 e2)
>   =  group (text "continue" <+> ppExpr e1 <+> ppExpr e2)
> ppExpr (Throw e)
>   =  group (text "throw" <+> ppExpr e)
> ppExpr (Try e rs)
>   =  group (text "try" <> block 2 (sp <> ppExpr e) <+> text "catch" <+>
>        block 2 (intersperse (sp <> text " | ") (map ppRule rs)) <+> text "end")
> ppExpr (Call f e)
>   =  group (pretty f <+> ppExpr e)
> ppExpr (While e1 e2)
>   =  group (text "while " <> 
>             block 2 (ppExpr e1 <> text " do" <+> ppExpr e2))

> ppExpr (Push e)
>   =  group (text "push" <+> ppExpr e)
> ppExpr (Pop e)
>   =  group (text "pop" <+> ppExpr e)
> ppExpr Struct
>   =  text "struct"
> ppExpr (LetOpen e1 e2)
>   =  group (text "let open " <> block 9 (ppExpr e1) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e2)))
> ppExpr (Lookup e x)
>   =  group (ppExpr e <> text "$" <> text x)

> ppExpr (Class e)
>   =  group (text "class" <+> ppExpr e <+> text "end")
> ppExpr (New e)
>   =  group (text "new" <+> ppExpr e)
> ppExpr Empty
>   =  text ""
> ppExpr (Append e1 e2)
>   =  group (ppExpr e1 <+> ppExpr e2)
> ppExpr (Method x e)
>   =  parens (text "method" <+> text x <> text " =" <+> ppExpr e)
> ppExpr (Inherit e)
>   =  group (text "inherit" <+> ppExpr e)
> ppExpr (InheritIn x e1 e2)
>   =  group (text "inherit " <> block 8 (text x <> text " =" <+> ppExpr e1) <> 
>             block 1 (sp <> text "in " <> block 3 (ppExpr e2) <+> text "end"))

 ppExpr (Mth p e)
   =  parens (align "method " (ppPat p <> text " =>" <+> ppExpr e))
 ppExpr Object
   =  text "end"
 ppExpr (Delay e)
   =  group (text "delay" <+> ppExpr e)
 ppExpr (Force e)
   =  group (text "force" <+> ppExpr e)
 ppExpr (LetInherit Nothing e1 e2)
   =  group (text "let inherit " <> block 9 (ppExpr e1) <> 
             block 1 (sp <> text "in " <> block 3 (ppExpr e2)))
 ppExpr (LetInherit (Just x) e1 e2)
   =  group (text "let inherit " <> text x <+> text "=" <+> block 9 (ppExpr e1) <> 
             block 1 (sp <> text "in " <> block 3 (ppExpr e2)))

> ppExpr (ArrFun e1 x e2)
>   =  parens (align "array " (brackets (ppExpr e1) <+> text x <> text " =>" <+> ppExpr e2))
> ppExpr (ArrList es)
>   =  ppList ppExpr es
> ppExpr (Subscript e1 e2)
>   =  group (ppExpr e1 <> brackets (ppExpr e2))

 ppExpr e -- catch-all (vorlufig)
   =  pretty e

> ppRule (p, e)
>   =  group (ppPat p <> text " => " <> ppExpr e)

> align s d = block (length s) (text s <> d)

> ppList pp as
>   =  brackets (intersperse (char ',' <> sp) (map pp as))

> infixl $
> ($) = App

Records.

> data Row a
>   =  Named     String a
>   |  Numbered  a
>      deriving (Show)

> instance Eq (Row a) where
>   Named s1 _  == Named s2 _  =  s1 == s2
>   Numbered _  == Numbered _  =  True
>   _           == _           =  False
>
> instance Ord (Row a) where
>   compare (Named s1 _) (Named s2 _)  =  compare s1 s2
>   compare (Named _ _ ) (Numbered _)  =  LT
>   compare (Numbered _) (Named _ _ )  =  GT
>   compare (Numbered _) (Numbered _)  =  EQ

> ppRow pp (Named s a)
>   =  align (s ++ " = ") (pp a)
> ppRow pp (Numbered a)
>   =  pp a

Objects and classes.

 data Slot
   =  SlMethod Var Expr
   |  SlLocal [Decl] [Slot]
   |  SlInherit Expr
   |  SlInheritIn Var Expr [Slot]
      deriving (Show)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Patterns}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Pat
>   =  PatFail                                          -- 0
>   |  PatSucceed                                       -- 1
>   |  PatVar     Var
>   |  PatCon     Con Pat
>   |  PatType                                          -- type abstraction marker
>   |  PatRecord  [Row Pat]
>   |  PatBoth    Pat Pat                               -- &
>   |  PatAlt     Pat Pat                               -- |
>      deriving (Show)

> patUnit       =  PatRecord []
> patRecord rs  =  PatRecord (sort rs)

NB. `sort' must be a stable sorting algorithm, so that the relative
order of numbered fields is not changed.

> ppPat PatFail
>   =  text "fail"
> ppPat PatSucceed
>   =  text "_"
> ppPat (PatVar x)
>   =  text x
> ppPat (PatCon c p)
>   =  parens (text c <+> ppPat p)
> ppPat PatType
>   =  text "<>"
> ppPat (PatRecord rs)
>   =  parens (group (intersperse (char ',' <> sp) (map (ppRow ppPat) rs)))
> ppPat (PatAlt p1 p2)
>   =  parens (group (ppPat p1 <> text " " <> block 2 (text "|" <+> ppPat p2)))
> ppPat (PatBoth p1 p2)
>   =  parens (group (ppPat p1 <> text " " <> block 2 (text "&" <+> ppPat p2)))

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{AST to core}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> core :: AS.Expr -> Expr
> core (AS.Bool _ b) 
>   =  Bool b
> core (AS.If e1 e2 e3) 
>   =  If (core e1) (core e2) (core e3)
> core (AS.UnOp op e) 
>   =  UnOp op (core e)
> core (AS.BinOp e1 Or e2)
>   =  If (core e1) (Bool True) (core e2)
> core (AS.BinOp e1 And e2)
>   =  If (core e1) (core e2) (Bool False)
> core (AS.BinOp e1 op e2) 
>   =  BinOp (core e1) op (core e2)
> core (AS.Nat _ n)
>   =  Nat n
> core (AS.String _ s) 
>   =  String s
> core (AS.Var (_,x)) 
>   =  Var x
> core (AS.Let ds e) 
>   =  coreLet ds (core e)
> core (AS.Fun xs e)
>   =  coreAbs xs (core e)
> core (AS.RecFun (_,f) xs t e)
>   =  Rec (PatVar f) (coreAbs xs (core e))
> core (AS.App (AS.QualCon _ (_,c)) [AS.ActExpr e])         -- unary constructor
>   =  Con c (core e)
> core (AS.App (AS.Con (_,c)) [AS.ActExpr e])               -- unary constructor
>   =  Con c (core e)
> core (AS.App e es) 
>   =  coreApp (core e) es
> core (AS.Rec p e) 
>   =  Rec (corePat p) (core e)
> core (AS.Record rs) 
>   =  Record (map (coreRow core) rs)
> core (AS.Con (_,c))
>   =  Con c unit                                           -- nullary constructor

> core (AS.Sel e l) = Sel (core e) l

> core (AS.Seq e1 e2) = Seq (core e1) (core e2)
> core (AS.Assign e1 e2) = Assign (core e1) (core e2)
> core (AS.Ref e) = Ref (core e)
> core (AS.Deref e) = Deref (core e)

> core (AS.Continue e1 e2)
>   =  Continue (core e1) (core e2)
> core (AS.LetCC (_,x) t e)
>   =  LetCC x (core e)

> core (AS.Case e rs) = Case (core e) (map coreRule rs)
> core (AS.Throw e) = Throw (core e)
> core (AS.Try e rs) = Try (core e) (map coreRule rs)
> -- core (AS.Call f e) = Call f (core e)
> core (AS.While e1 e2) = While (core e1) (core e2)
> core (AS.QualVar e (_,x)) = Lookup (coreStr e) x
> core (AS.QualCon e (_,c)) = Con c (unit)                  -- nullary constructor !! e wird ignoriert

> core (AS.Class ss)
>   =  Class (coreList coreSlot ss)
> core (AS.New e) 
>   =  New (core e)
> core (AS.ArrFun e1 (_,x) e2) 
>   =  ArrFun (core e1) x (core e2)
> core (AS.ArrList es) 
>   =  ArrList (map core es)
> core (AS.Subscript e1 e2) 
>   =  Subscript (core e1) (core e2)
> core (AS.Typed e t)
>   =  core e

> core e = error (show e)

> coreList core = foldr append Empty . map core

> coreRule (p, e)
>   =  (corePat p, core e)

> corePat :: AS.Pat -> Pat
> corePat AS.PatSucceed
>   =  PatSucceed
> corePat (AS.PatVar (_,x)) 
>   =  PatVar x
> corePat (AS.PatCon (_,c) Nothing) 
>   =  PatCon c patUnit
> corePat (AS.PatCon (_,c) (Just p)) 
>   =  PatCon c (corePat p)
> corePat (AS.PatTyped p t) 
>   =  corePat p
> corePat (AS.PatRecord rs)
>   =  patRecord (map (coreRow corePat) rs)
> corePat (AS.PatAlt p1 p2) 
>   =  PatAlt (corePat p1) (corePat p2)
> corePat (AS.PatBoth p1 p2) 
>   =  PatBoth (corePat p1) (corePat p2)

> coreAbs [] e 
>   =  e

> coreAbs (AS.ForPat p : xs) e
>   =  Fun (corePat p) (coreAbs xs e)
> coreAbs (AS.ForTypePat [] : xs) e
>   =  coreAbs xs e
> coreAbs (AS.ForTypePat (a : as) : xs) e
>   =  Fun PatType (coreAbs (AS.ForTypePat as : xs) e)

Die Typabstraktion `\ <a> => e' wird nach `\ () => e' berfhrt; die
Typapplikation `e <t>' entsprechend nach `e ()'. Das ist notwendig, um
die Auswertung zu verzgern (\ <a> => ref (raise Undefined)), siehe
auch Pierce, Types and Programming Languages, Abschnitt 23.7.
Sprachen mit implizitem Polymorphismus mssen z.B. die Form von
polymorphen Ausdrcken einschrnken (value restriction in Standard ML).

> coreApp e [] 
>   =  e
> coreApp e1 (AS.ActExpr e2 : es) 
>   =  coreApp (App e1 (core e2)) es
> coreApp e1 (AS.ActType [] : es) 
>   =  coreApp e1 es
> coreApp e1 (AS.ActType (t : ts) : es) 
>   =  coreApp (App e1 Type) (AS.ActType ts : es)

> coreLet [] e'
>   =  e'
> coreLet (AS.Val p1 e1 : ds) e'
>   =  Let (corePat p1) (core e1) (coreLet ds e')
> coreLet (AS.Function [((_,f), xs, t, e)] : ds) e'
>   =  Let (PatVar f) (Rec (PatVar f) (coreAbs xs (core e))) (coreLet ds e')
> coreLet (AS.Function fs : ds) e'
>   =  Let p (Rec p e) (coreLet ds e')
>      where p = patRecord [ Named f (PatVar f)            | ((_,f), xs, t, e) <- fs ]
>            e = Record    [ Named f (coreAbs xs (core e)) | ((_,f), xs, t, e) <- fs ]
> coreLet (AS.Type _ _ _ : ds) e'
>   =  coreLet ds e'
> coreLet (AS.Datatype _ _ _ : ds) e'
>   =  coreLet ds e'
> coreLet (AS.Data _ : ds) e'
>   =  coreLet ds e'
> coreLet (AS.Constr _ _ : ds) e'
>   =  coreLet ds e'
> coreLet (AS.Exception _ : ds) e'
>   =  coreLet ds e'
> coreLet (AS.Module (_,x1) fs e1 : ds) e'
>   =  Let (PatVar x1) (coreStr (AS.Functor fs e1)) (coreLet ds e')
> coreLet (AS.Open e : ds) e'
>   =  LetOpen (coreStr e) (coreLet ds e')
> coreLet (AS.Local ds1 ds2 : ds) e'
>   =  Push (coreLet ds1 (Push (coreLet ds2 (Pop (coreLet ds e')))))
> coreLet (AS.Interface _ _ _ : ds) e'
>   =  coreLet ds e'

 coreLet (AS.Class c xs e : ds) e'
   =  Let (PatVar c) (coreAbs xs (coreCls e)) (coreLet ds e')

 coreLet (AS.Inherit x e es : ds) e'
   =  LetInherit x (coreApp (Force (coreCls e)) es) (coreLet ds e')
 coreLet (AS.Method m xs t e : ds) e'
   =  Let (PatVar m) (Mth p1 e1) (coreLet ds e')
      where Fun p1 e1 =  coreAbs xs (core e)

> coreRow c (AS.Field Nothing e)
>   =  Numbered (c e)
> coreRow c (AS.Field (Just (AlNumLab l)) e)
>   =  Named l (c e)

> coreStr (AS.StrVar (_,x))
>   =  Var x
> coreStr (AS.StrSel e (_,x))
>   =  Lookup (coreStr e) x
> coreStr (AS.Struct ds)
>   =  Push (coreLet ds Struct)
> coreStr (AS.Functor fs e)
>   =  coreStrAbs fs (coreStr e)
> coreStr (AS.StrApp e as)
>   =  coreStrApp (coreStr e) as

> coreStrApp e []
>   =  e
> coreStrApp e1 (AS.StrActStrExpr e2 : es)
>   =  coreStrApp (App e1 (coreStr e2)) es
> coreStrApp e1 (AS.StrActType [] : es)
>   =  coreStrApp e1 es
> coreStrApp e1 (AS.StrActType (t : ts) : es) 
>   =  coreStrApp (App e1 Type) (AS.StrActType ts : es)

> coreStrAbs [] e
>   =  e
> coreStrAbs (AS.StrForStrPat (_,x) t : xs) e
>   =  Fun (PatVar x) (coreStrAbs xs e)
> coreStrAbs (AS.StrForTypePat [] : xs) e
>   =  coreStrAbs xs e
> coreStrAbs (AS.StrForTypePat (a : as) : xs) e
>   =  Fun PatType (coreStrAbs (AS.StrForTypePat as : xs) e)

 coreCls (AS.ClsVar x)
   =  Var x
 coreCls (AS.Object ds)
   =  Delay (Push (coreLet ds Object))
 coreCls (AS.ClsApp e as)
   =  coreApp (coreCls e) as
 coreCls (AS.ClsFun xs e)
   =  coreAbs xs (coreCls e)
 coreCls (AS.ClsLet ds e)
   =  coreLet ds (coreCls e)

> coreSlot (AS.SlMethod (_,m) fs t e)
>  =  Method m (coreAbs fs (core e))
> coreSlot (AS.SlLocal ds ss)
>  =  coreLet ds (coreList coreSlot ss)
> coreSlot (AS.SlInherit e)
>  =  Inherit (core e)
> coreSlot (AS.SlInheritIn (_,x) e ss)
>  =  InheritIn x (core e) (coreList coreSlot ss)

> expr2core :: AS.Expr -> Pass Expr
> expr2core e
>   =  let c = core e in
>      Succeed c  [Subtitle (text "abstract syntax tree (expr) -> core"),
>                  Debug (text "core"),
>                  Debug (pretty c)]

> decls2core :: [AS.Decl] -> Pass (Expr -> Expr)
> decls2core ds
>   =  let c = coreLet ds in
>      Succeed c  [Subtitle (text "abstract syntax tree (module) -> core"),
>                  Debug (text "core"),
>                  Debug (pretty (c Struct))]
